@@ -69,6 +69,7 @@ enum {
 };
 
 enum {
+        BEGIN_TIMED_LOGIN,
         BEGIN_VERIFICATION,
         BEGIN_VERIFICATION_FOR_USER,
         QUERY_ANSWER,
@@ -221,6 +222,20 @@ gdm_greeter_server_selected_user_changed (GdmGreeterServer *greeter_server,
         send_dbus_string_signal (greeter_server, "SelectedUserChanged", username);
 }
 
+void
+gdm_greeter_server_default_language_name_changed (GdmGreeterServer *greeter_server,
+                                                  const char       *language_name)
+{
+        send_dbus_string_signal (greeter_server, "DefaultLanguageNameChanged", language_name);
+}
+
+void
+gdm_greeter_server_default_session_name_changed (GdmGreeterServer *greeter_server,
+                                                 const char       *session_name)
+{
+        send_dbus_string_signal (greeter_server, "DefaultSessionNameChanged", session_name);
+}
+
 /* Note: Use abstract sockets like dbus does by default on Linux. Abstract
  * sockets are only available on Linux.
  */
@@ -267,6 +282,23 @@ handle_begin_verification (GdmGreeterServer *greeter_server,
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
+static DBusHandlerResult
+handle_begin_timed_login (GdmGreeterServer *greeter_server,
+                        DBusConnection   *connection,
+                        DBusMessage      *message)
+{
+        DBusMessage *reply;
+
+        g_debug ("GreeterServer: BeginTimedLogin");
+
+        reply = dbus_message_new_method_return (message);
+        dbus_connection_send (connection, reply, NULL);
+        dbus_message_unref (reply);
+
+        g_signal_emit (greeter_server, signals [BEGIN_TIMED_LOGIN], 0);
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
 
 static DBusHandlerResult
 handle_begin_verification_for_user (GdmGreeterServer *greeter_server,
@@ -490,13 +522,15 @@ greeter_handle_child_message (DBusConnection *connection,
                 return handle_begin_verification (greeter_server, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "BeginVerificationForUser")) {
                 return handle_begin_verification_for_user (greeter_server, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "BeginTimedLogin")) {
+                return handle_begin_timed_login (greeter_server, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "AnswerQuery")) {
                 return handle_answer_query (greeter_server, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "SelectSession")) {
                 return handle_select_session (greeter_server, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "HostnameSession")) {
                 return handle_select_hostname (greeter_server, connection, message);
-        } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "SelectSession")) {
+        } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "SelectLanguage")) {
                 return handle_select_language (greeter_server, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_GREETER_SERVER_DBUS_INTERFACE, "SelectUser")) {
                 return handle_select_user (greeter_server, connection, message);
@@ -535,6 +569,7 @@ do_introspect (DBusConnection *connection,
         xml = g_string_append (xml,
                                "  <interface name=\"org.gnome.DisplayManager.GreeterServer\">\n"
                                "    <method name=\"BeginVerification\">\n"
+                               "    <method name=\"BeginTimedLogin\">\n"
                                "    </method>\n"
                                "    <method name=\"BeginVerificationForUser\">\n"
                                "      <arg name=\"username\" direction=\"in\" type=\"s\"/>\n"
@@ -576,6 +611,12 @@ do_introspect (DBusConnection *connection,
                                "    <signal name=\"SelectedUserChanged\">\n"
                                "      <arg name=\"username\" type=\"s\"/>\n"
                                "    </signal>\n"
+                               "    <signal name=\"DefaultLanguageNameChanged\">\n"
+                               "      <arg name=\"language_name\" type=\"s\"/>\n"
+                               "    </signal>\n"
+                               "    <signal name=\"DefaultSessionNameChanged\">\n"
+                               "      <arg name=\"session_name\" type=\"s\"/>\n"
+                               "    </signal>\n"
                                "    <signal name=\"Ready\">\n"
                                "    </signal>\n"
                                "    <signal name=\"Reset\">\n"
@@ -896,9 +937,6 @@ gdm_greeter_server_constructor (GType                  type,
                                GObjectConstructParam *construct_properties)
 {
         GdmGreeterServer      *greeter_server;
-        GdmGreeterServerClass *klass;
-
-        klass = GDM_GREETER_SERVER_CLASS (g_type_class_peek (GDM_TYPE_GREETER_SERVER));
 
         greeter_server = GDM_GREETER_SERVER (G_OBJECT_CLASS (gdm_greeter_server_parent_class)->constructor (type,
                                                                                        n_construct_properties,
@@ -950,6 +988,16 @@ gdm_greeter_server_class_init (GdmGreeterServerClass *klass)
                               g_cclosure_marshal_VOID__VOID,
                               G_TYPE_NONE,
                               0);
+        signals [BEGIN_TIMED_LOGIN] =
+                g_signal_new ("begin-timed-login",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_FIRST,
+                              G_STRUCT_OFFSET (GdmGreeterServerClass, begin_timed_login),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE,
+                              0);
         signals [BEGIN_VERIFICATION_FOR_USER] =
                 g_signal_new ("begin-verification-for-user",
                               G_OBJECT_CLASS_TYPE (object_class),
