@@ -32,6 +32,7 @@
 #include <pwd.h>
 #include <grp.h>
 #include <signal.h>
+#include <sys/resource.h>
 
 #include <glib.h>
 #include <glib/gi18n.h>
@@ -99,6 +100,8 @@ enum {
 
 enum {
         READY,
+        EXITED,
+        DIED,
         LAST_SIGNAL
 };
 
@@ -574,6 +577,14 @@ server_child_watch (GPid       pid,
                  : WIFSIGNALED (status) ? WTERMSIG (status)
                  : -1);
 
+        if (WIFEXITED (status)) {
+                int code = WEXITSTATUS (status);
+                g_signal_emit (server, signals [EXITED], 0, code);
+        } else if (WIFSIGNALED (status)) {
+                int num = WTERMSIG (status);
+                g_signal_emit (server, signals [DIED], 0, num);
+        }
+
         g_spawn_close_pid (server->priv->pid);
         server->priv->pid = -1;
 }
@@ -742,6 +753,14 @@ _gdm_server_set_display_name (GdmServer  *server,
         server->priv->display_name = g_strdup (name);
 }
 
+static void
+_gdm_server_set_auth_file (GdmServer  *server,
+                           const char *auth_file)
+{
+        g_free (server->priv->auth_file);
+        server->priv->auth_file = g_strdup (auth_file);
+}
+
 static void
 _gdm_server_set_user_name (GdmServer  *server,
                            const char *name)
@@ -764,6 +783,9 @@ gdm_server_set_property (GObject      *object,
         case PROP_DISPLAY_NAME:
                 _gdm_server_set_display_name (self, g_value_get_string (value));
                 break;
+        case PROP_AUTH_FILE:
+                _gdm_server_set_auth_file (self, g_value_get_string (value));
+                break;
         case PROP_USER_NAME:
                 _gdm_server_set_user_name (self, g_value_get_string (value));
                 break;
@@ -791,6 +813,9 @@ gdm_server_get_property (GObject    *object,
                 g_value_take_string (value,
                                      gdm_server_get_display_device (self));
                 break;
+        case PROP_AUTH_FILE:
+                g_value_set_string (value, self->priv->auth_file);
+                break;
         case PROP_USER_NAME:
                 g_value_set_string (value, self->priv->user_name);
                 break;
@@ -806,9 +831,6 @@ gdm_server_constructor (GType                  type,
                         GObjectConstructParam *construct_properties)
 {
         GdmServer      *server;
-        GdmServerClass *klass;
-
-        klass = GDM_SERVER_CLASS (g_type_class_peek (GDM_TYPE_SERVER));
 
         server = GDM_SERVER (G_OBJECT_CLASS (gdm_server_parent_class)->constructor (type,
                                                                                     n_construct_properties,
@@ -838,6 +860,28 @@ gdm_server_class_init (GdmServerClass *klass)
                               g_cclosure_marshal_VOID__VOID,
                               G_TYPE_NONE,
                               0);
+        signals [EXITED] =
+                g_signal_new ("exited",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_FIRST,
+                              G_STRUCT_OFFSET (GdmServerClass, exited),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__INT,
+                              G_TYPE_NONE,
+                              1,
+                              G_TYPE_INT);
+        signals [DIED] =
+                g_signal_new ("died",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_FIRST,
+                              G_STRUCT_OFFSET (GdmServerClass, died),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__INT,
+                              G_TYPE_NONE,
+                              1,
+                              G_TYPE_INT);
 
         g_object_class_install_property (object_class,
                                          PROP_DISPLAY_NAME,
@@ -853,6 +897,13 @@ gdm_server_class_init (GdmServerClass *klass)
                                                               "Path to terminal display is running on",
                                                               NULL,
                                                               G_PARAM_READABLE));
+        g_object_class_install_property (object_class,
+                                         PROP_AUTH_FILE,
+                                         g_param_spec_string ("auth-file",
+                                                              "Authorization File",
+                                                              "Path to X authorization file",
+                                                              NULL,
+                                                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
         g_object_class_install_property (object_class,
                                          PROP_USER_NAME,
@@ -897,12 +948,14 @@ gdm_server_finalize (GObject *object)
 }
 
 GdmServer *
-gdm_server_new (const char *display_name)
+gdm_server_new (const char *display_name,
+                const char *auth_file)
 {
         GObject *object;
 
         object = g_object_new (GDM_TYPE_SERVER,
                                "display-name", display_name,
+                               "auth-file", auth_file,
                                NULL);
 
         return GDM_SERVER (object);
