@@ -17,10 +17,6 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  * 02111-1307, USA.
- *
- * TODO:                - close should be nicer and shutdown
- *                        pam etc
- *                      - audit libs (linux and solaris) support
  */
 
 #include "config.h"
@@ -39,11 +35,13 @@
 #include <pwd.h>
 #include <grp.h>
 
+#include <locale.h>
+
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
 #include <glib-object.h>
-#define DBUS_API_SUBJECT_TO_CHANGE
+
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 
@@ -64,7 +62,9 @@ struct _GdmSessionDirectPrivate
 {
         /* per open scope */
         char                *selected_session;
+        char                *saved_session;
         char                *selected_language;
+        char                *saved_language;
         char                *selected_user;
         char                *user_x11_authority_file;
 
@@ -78,10 +78,10 @@ struct _GdmSessionDirectPrivate
         guint32              is_running : 1;
 
         /* object lifetime scope */
-        char                *service_name;
         char                *display_name;
         char                *display_hostname;
         char                *display_device;
+        char                *display_x11_authority_file;
         gboolean             display_is_local;
 
         DBusServer          *server;
@@ -95,6 +95,7 @@ enum {
         PROP_DISPLAY_HOSTNAME,
         PROP_DISPLAY_IS_LOCAL,
         PROP_DISPLAY_DEVICE,
+        PROP_DISPLAY_X11_AUTHORITY_FILE,
         PROP_USER_X11_AUTHORITY_FILE,
 };
 
@@ -104,7 +105,7 @@ G_DEFINE_TYPE_WITH_CODE (GdmSessionDirect,
                          gdm_session_direct,
                          G_TYPE_OBJECT,
                          G_IMPLEMENT_INTERFACE (GDM_TYPE_SESSION,
-                                                gdm_session_iface_init));
+                                                gdm_session_iface_init))
 
 static gboolean
 send_dbus_message (DBusConnection *connection,
@@ -447,6 +448,35 @@ gdm_session_direct_handle_accreditation_failed (GdmSessionDirect *session,
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
+static const char *
+get_default_language_name (GdmSessionDirect *session)
+{
+    if (session->priv->saved_language != NULL) {
+                return session->priv->saved_language;
+    }
+
+    return setlocale (LC_MESSAGES, NULL);
+}
+
+static const char *
+get_default_session_name (GdmSessionDirect *session)
+{
+        if (session->priv->saved_session != NULL) {
+                return session->priv->saved_session;
+        }
+
+        return "gnome";
+}
+
+static void
+gdm_session_direct_defaults_changed (GdmSessionDirect *session)
+{
+        _gdm_session_default_language_name_changed (GDM_SESSION (session),
+                                                    get_default_language_name (session));
+        _gdm_session_default_session_name_changed (GDM_SESSION (session),
+                                                   get_default_session_name (session));
+}
+
 static void
 gdm_session_direct_select_user (GdmSession *session,
                                 const char *text)
@@ -457,6 +487,12 @@ gdm_session_direct_select_user (GdmSession *session,
 
         g_free (impl->priv->selected_user);
         impl->priv->selected_user = g_strdup (text);
+
+        g_free (impl->priv->saved_session);
+        impl->priv->saved_session = NULL;
+
+        g_free (impl->priv->saved_language);
+        impl->priv->saved_language = NULL;
 }
 
 static DBusHandlerResult
@@ -487,6 +523,8 @@ gdm_session_direct_handle_username_changed (GdmSessionDirect *session,
 
         _gdm_session_selected_user_changed (GDM_SESSION (session), session->priv->selected_user);
 
+        gdm_session_direct_defaults_changed (session);
+
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
@@ -771,6 +809,71 @@ gdm_session_direct_handle_session_died (GdmSessionDirect *session,
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
+static DBusHandlerResult
+gdm_session_direct_handle_saved_language_name_read (GdmSessionDirect *session,
+                                                    DBusConnection   *connection,
+                                                    DBusMessage      *message)
+{
+        DBusMessage *reply;
+        DBusError    error;
+        const char  *language_name;
+
+        dbus_error_init (&error);
+        if (! dbus_message_get_args (message, &error,
+                                     DBUS_TYPE_STRING, &language_name,
+                                     DBUS_TYPE_INVALID)) {
+                g_warning ("ERROR: %s", error.message);
+        }
+
+        reply = dbus_message_new_method_return (message);
+        dbus_connection_send (connection, reply, NULL);
+        dbus_message_unref (reply);
+
+        if (strcmp (language_name,
+                    get_default_language_name (session)) != 0) {
+                g_free (session->priv->saved_language);
+                session->priv->saved_language = g_strdup (language_name);
+
+                _gdm_session_default_language_name_changed (GDM_SESSION (session),
+                                                            language_name);
+        }
+
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
+
+static DBusHandlerResult
+gdm_session_direct_handle_saved_session_name_read (GdmSessionDirect *session,
+                                                   DBusConnection   *connection,
+                                                   DBusMessage      *message)
+{
+        DBusMessage *reply;
+        DBusError    error;
+        const char  *session_name;
+
+        dbus_error_init (&error);
+        if (! dbus_message_get_args (message, &error,
+                                     DBUS_TYPE_STRING, &session_name,
+                                     DBUS_TYPE_INVALID)) {
+                g_warning ("ERROR: %s", error.message);
+        }
+
+        reply = dbus_message_new_method_return (message);
+        dbus_connection_send (connection, reply, NULL);
+        dbus_message_unref (reply);
+
+        if (strcmp (session_name,
+                    get_default_session_name (session)) != 0) {
+                g_free (session->priv->saved_session);
+                session->priv->saved_session = g_strdup (session_name);
+
+                _gdm_session_default_session_name_changed (GDM_SESSION (session),
+                                                           session_name);
+        }
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
+
 static DBusHandlerResult
 session_worker_message (DBusConnection *connection,
                         DBusMessage    *message,
@@ -818,6 +921,10 @@ session_worker_message (DBusConnection *connection,
                 return gdm_session_direct_handle_session_exited (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SessionDied")) {
                 return gdm_session_direct_handle_session_died (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedLanguageNameRead")) {
+                return gdm_session_direct_handle_saved_language_name_read (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedSessionNameRead")) {
+                return gdm_session_direct_handle_saved_session_name_read (session, connection, message);
         }
 
         return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -909,12 +1016,14 @@ do_introspect (DBusConnection *connection,
                                "      <arg name=\"x11_display_name\" type=\"s\"/>\n"
                                "      <arg name=\"display_device\" type=\"s\"/>\n"
                                "      <arg name=\"hostname\" type=\"s\"/>\n"
+                               "      <arg name=\"x11_authority_file\" type=\"s\"/>\n"
                                "    </signal>\n"
                                "    <signal name=\"SetupForUser\">\n"
                                "      <arg name=\"service_name\" type=\"s\"/>\n"
                                "      <arg name=\"x11_display_name\" type=\"s\"/>\n"
                                "      <arg name=\"display_device\" type=\"s\"/>\n"
                                "      <arg name=\"hostname\" type=\"s\"/>\n"
+                               "      <arg name=\"x11_authority_file\" type=\"s\"/>\n"
                                "      <arg name=\"username\" type=\"s\"/>\n"
                                "    </signal>\n"
                                "    <signal name=\"Authenticate\">\n"
@@ -1138,8 +1247,6 @@ setup_server (GdmSessionDirect *session)
 static void
 gdm_session_direct_init (GdmSessionDirect *session)
 {
-        const char * const *languages;
-
         session->priv = G_TYPE_INSTANCE_GET_PRIVATE (session,
                                                      GDM_TYPE_SESSION_DIRECT,
                                                      GdmSessionDirectPrivate);
@@ -1161,14 +1268,7 @@ gdm_session_direct_init (GdmSessionDirect *session)
                           G_CALLBACK (on_session_exited),
                           NULL);
 
-        languages = g_get_language_names ();
-        if (languages != NULL) {
-                session->priv->selected_language = g_strdup (languages[0]);
-        }
-
         session->priv->session_pid = -1;
-        session->priv->selected_session = g_strdup ("gnome.desktop");
-        session->priv->service_name = g_strdup ("gdm");
 
         session->priv->environment = g_hash_table_new_full (g_str_hash,
                                                             g_str_equal,
@@ -1275,23 +1375,26 @@ static void
 gdm_session_direct_open (GdmSession *session)
 {
         GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
-        gboolean          res;
 
         g_return_if_fail (session != NULL);
 
         g_debug ("GdmSessionDirect: Opening session");
 
-        res = start_worker (impl);
+        start_worker (impl);
 }
 
 static void
-send_setup (GdmSessionDirect *session)
+send_setup (GdmSessionDirect *session,
+            const char       *service_name)
 {
         DBusMessage    *message;
         DBusMessageIter iter;
         const char     *display_name;
         const char     *display_device;
         const char     *display_hostname;
+        const char     *display_x11_authority_file;
+
+        g_assert (service_name != NULL);
 
         if (session->priv->display_name != NULL) {
                 display_name = session->priv->display_name;
@@ -1308,6 +1411,11 @@ send_setup (GdmSessionDirect *session)
         } else {
                 display_device = "";
         }
+        if (session->priv->display_x11_authority_file != NULL) {
+                display_x11_authority_file = session->priv->display_x11_authority_file;
+        } else {
+                display_x11_authority_file = "";
+        }
 
         g_debug ("GdmSessionDirect: Beginning setup");
 
@@ -1316,10 +1424,11 @@ send_setup (GdmSessionDirect *session)
                                            "Setup");
 
         dbus_message_iter_init_append (message, &iter);
-        dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &session->priv->service_name);
+        dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &service_name);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_name);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_device);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_hostname);
+        dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_x11_authority_file);
 
         if (! send_dbus_message (session->priv->worker_connection, message)) {
                 g_debug ("GdmSessionDirect: Could not send %s signal", "Setup");
@@ -1329,15 +1438,19 @@ send_setup (GdmSessionDirect *session)
 }
 
 static void
-send_setup_for_user (GdmSessionDirect *session)
+send_setup_for_user (GdmSessionDirect *session,
+                     const char       *service_name)
 {
         DBusMessage    *message;
         DBusMessageIter iter;
         const char     *display_name;
         const char     *display_device;
         const char     *display_hostname;
+        const char     *display_x11_authority_file;
         const char     *selected_user;
 
+        g_assert (service_name != NULL);
+
         if (session->priv->display_name != NULL) {
                 display_name = session->priv->display_name;
         } else {
@@ -1353,6 +1466,11 @@ send_setup_for_user (GdmSessionDirect *session)
         } else {
                 display_device = "";
         }
+        if (session->priv->display_x11_authority_file != NULL) {
+                display_x11_authority_file = session->priv->display_x11_authority_file;
+        } else {
+                display_x11_authority_file = "";
+        }
         if (session->priv->selected_user != NULL) {
                 selected_user = session->priv->selected_user;
         } else {
@@ -1366,10 +1484,11 @@ send_setup_for_user (GdmSessionDirect *session)
                                            "SetupForUser");
 
         dbus_message_iter_init_append (message, &iter);
-        dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &session->priv->service_name);
+        dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &service_name);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_name);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_device);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_hostname);
+        dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &display_x11_authority_file);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &selected_user);
 
         if (! send_dbus_message (session->priv->worker_connection, message)) {
@@ -1380,19 +1499,22 @@ send_setup_for_user (GdmSessionDirect *session)
 }
 
 static void
-gdm_session_direct_setup (GdmSession *session)
+gdm_session_direct_setup (GdmSession *session,
+                          const char *service_name)
 {
         GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
 
         g_return_if_fail (session != NULL);
         g_return_if_fail (dbus_connection_get_is_connected (impl->priv->worker_connection));
 
-        send_setup (impl);
+        send_setup (impl, service_name);
+        gdm_session_direct_defaults_changed (impl);
 }
 
 static void
-gdm_session_direct_setup_for_user (GdmSession  *session,
-                                   const char  *username)
+gdm_session_direct_setup_for_user (GdmSession *session,
+                                   const char *service_name,
+                                   const char *username)
 {
         GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
 
@@ -1402,7 +1524,8 @@ gdm_session_direct_setup_for_user (GdmSession  *session,
 
         gdm_session_direct_select_user (session, username);
 
-        send_setup_for_user (impl);
+        send_setup_for_user (impl, service_name);
+        gdm_session_direct_defaults_changed (impl);
 }
 
 static void
@@ -1518,6 +1641,8 @@ get_session_command_for_file (const char *file,
 
         key_file = g_key_file_new ();
 
+        g_debug ("GdmSessionDirect: looking for session file '%s'", file);
+
         error = NULL;
         full_path = NULL;
         res = g_key_file_load_from_dirs (key_file,
@@ -1587,19 +1712,43 @@ out:
         return ret;
 }
 
+static const char *
+get_language_name (GdmSessionDirect *session)
+{
+        if (session->priv->selected_language != NULL) {
+                return session->priv->selected_language;
+        }
+
+        return get_default_language_name (session);
+}
+
+static const char *
+get_session_name (GdmSessionDirect *session)
+{
+        if (session->priv->selected_session != NULL) {
+                return session->priv->selected_session;
+        }
+
+        return get_default_session_name (session);
+}
+
 static char *
 get_session_command (GdmSessionDirect *session)
 {
         gboolean res;
         char    *command;
+        char    *filename;
+
+        filename = g_strdup_printf ("%s.desktop", get_session_name (session));
 
         command = NULL;
-        res = get_session_command_for_file (session->priv->selected_session,
-                                            &command);
+        res = get_session_command_for_file (filename, &command);
         if (! res) {
-                g_critical ("Cannot read specified session file: %s", session->priv->selected_session);
+                g_critical ("Cannot read specified session file: %s", filename);
+                g_free (filename);
                 exit (1);
         }
+        g_free (filename);
 
         return command;
 }
@@ -1700,17 +1849,17 @@ setup_session_environment (GdmSessionDirect *session)
 
         gdm_session_direct_set_environment_variable (session,
                                                      "GDMSESSION",
-                                                     session->priv->selected_session);
+                                                     get_session_name (session));
         gdm_session_direct_set_environment_variable (session,
                                                      "DESKTOP_SESSION",
-                                                     session->priv->selected_session);
+                                                     get_session_name (session));
 
         gdm_session_direct_set_environment_variable (session,
                                                      "LANG",
-                                                     session->priv->selected_language);
+                                                     get_language_name (session));
         gdm_session_direct_set_environment_variable (session,
                                                      "GDM_LANG",
-                                                     session->priv->selected_language);
+                                                     get_language_name (session));
 
         gdm_session_direct_set_environment_variable (session,
                                                      "DISPLAY",
@@ -1787,9 +1936,15 @@ gdm_session_direct_close (GdmSession *session)
         g_free (impl->priv->selected_session);
         impl->priv->selected_session = NULL;
 
+        g_free (impl->priv->saved_session);
+        impl->priv->saved_session = NULL;
+
         g_free (impl->priv->selected_language);
         impl->priv->selected_language = NULL;
 
+        g_free (impl->priv->saved_language);
+        impl->priv->saved_language = NULL;
+
         g_free (impl->priv->user_x11_authority_file);
         impl->priv->user_x11_authority_file = NULL;
 
@@ -1836,7 +1991,12 @@ gdm_session_direct_select_session (GdmSession *session,
         GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
 
         g_free (impl->priv->selected_session);
-        impl->priv->selected_session = g_strdup (text);
+
+        if (strcmp (text, "__previous") == 0) {
+                impl->priv->selected_session = NULL;
+        } else {
+                impl->priv->selected_session = g_strdup (text);
+        }
 }
 
 static void
@@ -1846,7 +2006,12 @@ gdm_session_direct_select_language (GdmSession *session,
         GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
 
         g_free (impl->priv->selected_language);
-        impl->priv->selected_language = g_strdup (text);
+
+        if (strcmp (text, "__previous") == 0) {
+                impl->priv->selected_language = NULL;
+        } else {
+                impl->priv->selected_language = g_strdup (text);
+        }
 }
 
 /* At some point we may want to read these right from
@@ -1884,6 +2049,14 @@ _gdm_session_direct_set_user_x11_authority_file (GdmSessionDirect *session,
         session->priv->user_x11_authority_file = g_strdup (name);
 }
 
+static void
+_gdm_session_direct_set_display_x11_authority_file (GdmSessionDirect *session,
+                                                    const char       *name)
+{
+        g_free (session->priv->display_x11_authority_file);
+        session->priv->display_x11_authority_file = g_strdup (name);
+}
+
 static void
 _gdm_session_direct_set_display_is_local (GdmSessionDirect *session,
                                           gboolean          is)
@@ -1914,6 +2087,9 @@ gdm_session_direct_set_property (GObject      *object,
         case PROP_USER_X11_AUTHORITY_FILE:
                 _gdm_session_direct_set_user_x11_authority_file (self, g_value_get_string (value));
                 break;
+        case PROP_DISPLAY_X11_AUTHORITY_FILE:
+                _gdm_session_direct_set_display_x11_authority_file (self, g_value_get_string (value));
+                break;
         case PROP_DISPLAY_IS_LOCAL:
                 _gdm_session_direct_set_display_is_local (self, g_value_get_boolean (value));
                 break;
@@ -1946,6 +2122,9 @@ gdm_session_direct_get_property (GObject    *object,
         case PROP_USER_X11_AUTHORITY_FILE:
                 g_value_set_string (value, self->priv->user_x11_authority_file);
                 break;
+        case PROP_DISPLAY_X11_AUTHORITY_FILE:
+                g_value_set_string (value, self->priv->display_x11_authority_file);
+                break;
         case PROP_DISPLAY_IS_LOCAL:
                 g_value_set_boolean (value, self->priv->display_is_local);
                 break;
@@ -1966,9 +2145,6 @@ gdm_session_direct_dispose (GObject *object)
 
         gdm_session_direct_close (GDM_SESSION (session));
 
-        g_free (session->priv->service_name);
-        session->priv->service_name = NULL;
-
         g_free (session->priv->display_name);
         session->priv->display_name = NULL;
 
@@ -1978,6 +2154,9 @@ gdm_session_direct_dispose (GObject *object)
         g_free (session->priv->display_device);
         session->priv->display_device = NULL;
 
+        g_free (session->priv->display_x11_authority_file);
+        session->priv->display_x11_authority_file = NULL;
+
         g_free (session->priv->server_address);
         session->priv->server_address = NULL;
 
@@ -2004,6 +2183,10 @@ gdm_session_direct_finalize (GObject *object)
         session = GDM_SESSION_DIRECT (object);
 
         g_free (session->priv->selected_user);
+        g_free (session->priv->selected_session);
+        g_free (session->priv->saved_session);
+        g_free (session->priv->selected_language);
+        g_free (session->priv->saved_language);
 
         parent_class = G_OBJECT_CLASS (gdm_session_direct_parent_class);
 
@@ -2065,6 +2248,13 @@ gdm_session_direct_class_init (GdmSessionDirectClass *session_class)
                                                                "display is local",
                                                                TRUE,
                                                                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+        g_object_class_install_property (object_class,
+                                         PROP_DISPLAY_X11_AUTHORITY_FILE,
+                                         g_param_spec_string ("display-x11-authority-file",
+                                                              "display x11 authority file",
+                                                              "display x11 authority file",
+                                                              NULL,
+                                                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
         /* not construct only */
         g_object_class_install_property (object_class,
                                          PROP_USER_X11_AUTHORITY_FILE,
@@ -2088,6 +2278,7 @@ GdmSessionDirect *
 gdm_session_direct_new (const char *display_name,
                         const char *display_hostname,
                         const char *display_device,
+                        const char *display_x11_authority_file,
                         gboolean    display_is_local)
 {
         GdmSessionDirect *session;
@@ -2096,6 +2287,7 @@ gdm_session_direct_new (const char *display_name,
                                 "display-name", display_name,
                                 "display-hostname", display_hostname,
                                 "display-device", display_device,
+                                "display-x11-authority-file", display_x11_authority_file,
                                 "display-is-local", display_is_local,
                                 NULL);
 
