@@ -263,6 +263,7 @@ gdm_slave_run_script (GdmSlave   *slave,
 
         script = g_build_filename (dir, slave->priv->display_name, NULL);
         if (g_access (script, R_OK | X_OK) != 0) {
+                g_debug ("GdmSlave: script %s not found; skipping", script);
                 g_free (script);
                 script = NULL;
         }
@@ -271,6 +272,7 @@ gdm_slave_run_script (GdmSlave   *slave,
             slave->priv->display_hostname != NULL) {
                 script = g_build_filename (dir, slave->priv->display_hostname, NULL);
                 if (g_access (script, R_OK | X_OK) != 0) {
+                        g_debug ("GdmSlave: script %s not found; skipping", script);
                         g_free (script);
                         script = NULL;
                 }
@@ -298,12 +300,14 @@ gdm_slave_run_script (GdmSlave   *slave,
         if (script == NULL) {
                 script = g_build_filename (dir, "Default", NULL);
                 if (g_access (script, R_OK | X_OK) != 0) {
+                        g_debug ("GdmSlave: script %s not found; skipping", script);
                         g_free (script);
                         script = NULL;
                 }
         }
 
         if (script == NULL) {
+                g_debug ("GdmSlave: no script found");
                 return TRUE;
         }
 
@@ -333,6 +337,11 @@ gdm_slave_run_script (GdmSlave   *slave,
         g_ptr_array_foreach (env, (GFunc)g_free, NULL);
         g_ptr_array_free (env, TRUE);
 
+        if (! res) {
+                g_warning ("GdmSlave: Unable to run script: %s", error->message);
+                g_error_free (error);
+        }
+
         gdm_slave_whack_temp_auth_file (slave);
 
         if (WIFEXITED (status)) {
@@ -380,13 +389,6 @@ gdm_slave_connect_to_x11_display (GdmSlave *slave)
         g_setenv ("DISPLAY", slave->priv->display_name, TRUE);
         g_setenv ("XAUTHORITY", slave->priv->display_x11_authority_file, TRUE);
 
-#if 0
-        /* X error handlers to avoid the default one (i.e. exit (1)) */
-        do_xfailed_on_xio_error = TRUE;
-        XSetErrorHandler (gdm_slave_xerror_handler);
-        XSetIOErrorHandler (gdm_slave_xioerror_handler);
-#endif
-
         sigemptyset (&mask);
         sigaddset (&mask, SIGCHLD);
         sigprocmask (SIG_BLOCK, &mask, &omask);
@@ -685,6 +687,8 @@ gdm_slave_add_user_authorization (GdmSlave   *slave,
                 *filenamep = NULL;
         }
 
+        g_debug ("GdmSlave: Requesting user authorization");
+
         error = NULL;
         res = dbus_g_proxy_call (slave->priv->display_proxy,
                                  "AddUserAuthorization",
@@ -693,10 +697,6 @@ gdm_slave_add_user_authorization (GdmSlave   *slave,
                                  G_TYPE_INVALID,
                                  G_TYPE_STRING, &filename,
                                  G_TYPE_INVALID);
-        if (filenamep != NULL) {
-                *filenamep = g_strdup (filename);
-        }
-        g_free (filename);
 
         if (! res) {
                 if (error != NULL) {
@@ -705,8 +705,15 @@ gdm_slave_add_user_authorization (GdmSlave   *slave,
                 } else {
                         g_warning ("Failed to add user authorization");
                 }
+        } else {
+                g_debug ("GdmSlave: Got user authorization: %s", filename);
         }
 
+        if (filenamep != NULL) {
+                *filenamep = g_strdup (filename);
+        }
+        g_free (filename);
+
         return res;
 }
 
@@ -839,7 +846,6 @@ _get_primary_user_session_id (GdmSlave   *slave,
                               const char *username)
 {
         gboolean    res;
-        gboolean    ret;
         gboolean    can_activate_sessions;
         GError     *error;
         DBusGProxy *manager_proxy;
@@ -851,10 +857,9 @@ _get_primary_user_session_id (GdmSlave   *slave,
 
         if (slave->priv->display_seat_id == NULL || slave->priv->display_seat_id[0] == '\0') {
                 g_debug ("GdmSlave: display seat id is not set; can't switch sessions");
-                return FALSE;
+                return NULL;
         }
 
-        ret = FALSE;
         manager_proxy = NULL;
         primary_ssid = NULL;
         sessions = NULL;
@@ -917,18 +922,13 @@ _get_primary_user_session_id (GdmSlave   *slave,
 
                 ssid = g_ptr_array_index (sessions, i);
 
-                if (! x11_session_is_on_seat (slave, ssid, slave->priv->display_seat_id)) {
-                        goto next;
-                }
-
-                /* FIXME: better way to choose? */
-                if (primary_ssid == NULL) {
+                if (x11_session_is_on_seat (slave, ssid, slave->priv->display_seat_id)) {
                         primary_ssid = g_strdup (ssid);
+                        break;
                 }
-
-        next:
-                g_free (ssid);
         }
+
+        g_ptr_array_foreach (sessions, (GFunc)g_free, NULL);
         g_ptr_array_free (sessions, TRUE);
 
  out:
@@ -989,6 +989,40 @@ activate_session_id (GdmSlave   *slave,
         return ret;
 }
 
+static gboolean
+session_unlock (GdmSlave   *slave,
+                const char *ssid)
+{
+        DBusError       error;
+        DBusMessage    *message;
+        DBusMessage    *reply;
+
+        g_debug ("ConsoleKit: Unlocking session %s", ssid);
+        message = dbus_message_new_method_call (CK_NAME,
+                                                ssid,
+                                                CK_SESSION_INTERFACE,
+                                                "Unlock");
+        if (message == NULL) {
+                g_debug ("GdmSlave: ConsoleKit couldn't allocate the D-Bus message");
+                return FALSE;
+        }
+
+        dbus_error_init (&error);
+        reply = dbus_connection_send_with_reply_and_block (dbus_g_connection_get_connection (slave->priv->connection),
+                                                           message,
+                                                           -1, &error);
+        dbus_message_unref (message);
+        dbus_message_unref (reply);
+        dbus_connection_flush (dbus_g_connection_get_connection (slave->priv->connection));
+
+        if (dbus_error_is_set (&error)) {
+                g_debug ("GdmSlave: ConsoleKit %s raised:\n %s\n\n", error.name, error.message);
+                return FALSE;
+        }
+
+        return TRUE;
+}
+
 gboolean
 gdm_slave_switch_to_user_session (GdmSlave   *slave,
                                   const char *username)
@@ -1013,6 +1047,12 @@ gdm_slave_switch_to_user_session (GdmSlave   *slave,
                 goto out;
         }
 
+        res = session_unlock (slave, ssid_to_activate);
+        if (!res) {
+                /* this isn't fatal */
+                g_debug ("GdmSlave: unable to unlock session: %s", ssid_to_activate);
+        }
+
         ret = TRUE;
 
  out:
@@ -1177,12 +1217,9 @@ gdm_slave_constructor (GType                  type,
                        GObjectConstructParam *construct_properties)
 {
         GdmSlave      *slave;
-        GdmSlaveClass *klass;
         gboolean       res;
         const char    *id;
 
-        klass = GDM_SLAVE_CLASS (g_type_class_peek (GDM_TYPE_SLAVE));
-
         slave = GDM_SLAVE (G_OBJECT_CLASS (gdm_slave_parent_class)->constructor (type,
                                                                                  n_construct_properties,
                                                                                  construct_properties));
