@@ -22,6 +22,7 @@
 
 #include <stdlib.h>
 #include <stdio.h>
+#include <sys/time.h>
 #include <unistd.h>
 #include <string.h>
 
@@ -30,7 +31,12 @@
 #include <glib-object.h>
 #include <gtk/gtk.h>
 
+#include "gdm-languages.h"
 #include "gdm-greeter-panel.h"
+#include "gdm-clock-widget.h"
+#include "gdm-language-option-widget.h"
+#include "gdm-session-option-widget.h"
+#include "gdm-a11y-preferences-dialog.h"
 
 #include "na-tray.h"
 
@@ -40,14 +46,30 @@ struct GdmGreeterPanelPrivate
 {
         int                     monitor;
         GdkRectangle            geometry;
+        GtkWidget              *a11y_button;
+        GtkWidget              *a11y_dialog;
         GtkWidget              *hbox;
         GtkWidget              *hostname_label;
+        GtkWidget              *clock;
+        GtkWidget              *language_option_widget;
+        GtkWidget              *session_option_widget;
+
+        char                   *default_session_name;
+        char                   *default_language_name;
 };
 
 enum {
         PROP_0,
 };
 
+enum {
+        LANGUAGE_SELECTED,
+        SESSION_SELECTED,
+        NUMBER_OF_SIGNALS
+};
+
+static guint signals [NUMBER_OF_SIGNALS] = { 0, };
+
 static void     gdm_greeter_panel_class_init  (GdmGreeterPanelClass *klass);
 static void     gdm_greeter_panel_init        (GdmGreeterPanel      *greeter_panel);
 static void     gdm_greeter_panel_finalize    (GObject              *object);
@@ -60,10 +82,6 @@ gdm_greeter_panel_set_property (GObject        *object,
                                 const GValue   *value,
                                 GParamSpec     *pspec)
 {
-        GdmGreeterPanel *self;
-
-        self = GDM_GREETER_PANEL (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -77,10 +95,6 @@ gdm_greeter_panel_get_property (GObject        *object,
                                 GValue         *value,
                                 GParamSpec     *pspec)
 {
-        GdmGreeterPanel *self;
-
-        self = GDM_GREETER_PANEL (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -94,9 +108,6 @@ gdm_greeter_panel_constructor (GType                  type,
                                GObjectConstructParam *construct_properties)
 {
         GdmGreeterPanel      *greeter_panel;
-        GdmGreeterPanelClass *klass;
-
-        klass = GDM_GREETER_PANEL_CLASS (g_type_class_peek (GDM_TYPE_GREETER_PANEL));
 
         greeter_panel = GDM_GREETER_PANEL (G_OBJECT_CLASS (gdm_greeter_panel_parent_class)->constructor (type,
                                                                                                          n_construct_properties,
@@ -108,12 +119,6 @@ gdm_greeter_panel_constructor (GType                  type,
 static void
 gdm_greeter_panel_dispose (GObject *object)
 {
-        GdmGreeterPanel *greeter_panel;
-
-        greeter_panel = GDM_GREETER_PANEL (object);
-
-        g_debug ("Disposing greeter_panel");
-
         G_OBJECT_CLASS (gdm_greeter_panel_parent_class)->dispose (object);
 }
 
@@ -250,7 +255,9 @@ set_struts (GdmGreeterPanel *panel,
         /* bottom_end_x */
         data[11] = x + width;
 
+#if 0
         g_debug ("Setting strut: bottom=%lu bottom_start_x=%lu bottom_end_x=%lu", data[3], data[10], data[11]);
+#endif
 
         gdk_error_trap_push ();
 
@@ -293,22 +300,22 @@ update_geometry (GdmGreeterPanel *panel,
                  GtkRequisition  *requisition)
 {
         GdkRectangle geometry;
-        int          height;
 
         get_monitor_geometry (panel, &geometry);
 
-        height = requisition->height;
         panel->priv->geometry.width = geometry.width;
-        panel->priv->geometry.height = height * 2;
+        panel->priv->geometry.height = requisition->height + 2 * GTK_CONTAINER (panel)->border_width;
 
         panel->priv->geometry.x = geometry.x;
         panel->priv->geometry.y = geometry.y + geometry.height - panel->priv->geometry.height;
 
+#if 0
         g_debug ("Setting geometry x:%d y:%d w:%d h:%d",
                  panel->priv->geometry.x,
                  panel->priv->geometry.y,
                  panel->priv->geometry.width,
                  panel->priv->geometry.height);
+#endif
 
         update_struts (panel);
 }
@@ -369,40 +376,143 @@ gdm_greeter_panel_class_init (GdmGreeterPanelClass *klass)
         widget_class->realize = gdm_greeter_panel_real_realize;
         widget_class->unrealize = gdm_greeter_panel_real_unrealize;
         widget_class->size_request = gdm_greeter_panel_real_size_request;
+        signals[LANGUAGE_SELECTED] =
+                g_signal_new ("language-selected",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GdmGreeterPanelClass, language_selected),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__STRING,
+                              G_TYPE_NONE,
+                              1, G_TYPE_STRING);
+
+        signals[SESSION_SELECTED] =
+                g_signal_new ("session-selected",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GdmGreeterPanelClass, session_selected),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__STRING,
+                              G_TYPE_NONE,
+                              1, G_TYPE_STRING);
 
         g_type_class_add_private (klass, sizeof (GdmGreeterPanelPrivate));
 }
 
+static void
+on_language_activated (GdmLanguageOptionWidget *widget,
+                       GdmGreeterPanel         *panel)
+{
+
+        char *language;
+
+        language = gdm_language_option_widget_get_current_language_name (GDM_LANGUAGE_OPTION_WIDGET (panel->priv->language_option_widget));
+
+        if (language == NULL) {
+                return;
+        }
+
+        g_signal_emit (panel, signals[LANGUAGE_SELECTED], 0, language);
+
+        g_free (language);
+}
+
+static void
+on_session_activated (GdmSessionOptionWidget *widget,
+                      GdmGreeterPanel        *panel)
+{
+
+        char *session;
+
+        session = gdm_session_option_widget_get_current_session (GDM_SESSION_OPTION_WIDGET (panel->priv->session_option_widget));
+
+        if (session == NULL) {
+                return;
+        }
+
+        g_signal_emit (panel, signals[SESSION_SELECTED], 0, session);
+
+        g_free (session);
+}
+
+static void
+on_a11y_dialog_response (GtkDialog       *dialog,
+                         int              response,
+                         GdmGreeterPanel *panel)
+{
+        g_signal_handlers_disconnect_by_func (dialog,
+                                              on_a11y_dialog_response,
+                                              panel);
+
+        gtk_widget_destroy (GTK_WIDGET (dialog));
+        panel->priv->a11y_dialog = NULL;
+}
+
+static void
+on_a11y_button_clicked (GtkButton       *button,
+                        GdmGreeterPanel *panel)
+{
+        if (panel->priv->a11y_dialog == NULL) {
+                panel->priv->a11y_dialog = gdm_a11y_preferences_dialog_new ();
+                g_signal_connect (panel->priv->a11y_dialog,
+                                  "response",
+                                  G_CALLBACK (on_a11y_dialog_response),
+                                  panel);
+        }
+        gtk_window_present (GTK_WINDOW (panel->priv->a11y_dialog));
+}
+
 static void
 gdm_greeter_panel_init (GdmGreeterPanel *panel)
 {
-        GtkWidget *label;
         NaTray    *tray;
+        GtkWidget *image;
 
         panel->priv = GDM_GREETER_PANEL_GET_PRIVATE (panel);
 
+        GTK_WIDGET_SET_FLAGS (GTK_WIDGET (panel), GTK_CAN_FOCUS);
+
         panel->priv->geometry.x      = -1;
         panel->priv->geometry.y      = -1;
         panel->priv->geometry.width  = -1;
         panel->priv->geometry.height = -1;
 
+        gtk_window_set_title (GTK_WINDOW (panel), _("Panel"));
         gtk_window_set_decorated (GTK_WINDOW (panel), FALSE);
 
-        gtk_window_set_skip_taskbar_hint (GTK_WINDOW (panel), TRUE);
-        gtk_window_set_skip_pager_hint (GTK_WINDOW (panel), TRUE);
-
         gtk_window_set_keep_above (GTK_WINDOW (panel), TRUE);
         gtk_window_set_type_hint (GTK_WINDOW (panel), GDK_WINDOW_TYPE_HINT_DOCK);
         gtk_window_set_opacity (GTK_WINDOW (panel), 0.75);
 
         panel->priv->hbox = gtk_hbox_new (FALSE, 12);
+        gtk_container_set_border_width (GTK_CONTAINER (panel->priv->hbox), 0);
         gtk_widget_show (panel->priv->hbox);
         gtk_container_add (GTK_CONTAINER (panel), panel->priv->hbox);
 
-        /* Add a label so that the panel gets sized correctly */
-        label = gtk_label_new ("");
-        gtk_box_pack_start (GTK_BOX (panel->priv->hbox), label, FALSE, FALSE, 6);
-        gtk_widget_show (label);
+        panel->priv->a11y_button = gtk_button_new ();
+        image = gtk_image_new_from_icon_name ("preferences-desktop-accessibility", GTK_ICON_SIZE_BUTTON);
+        gtk_container_add (GTK_CONTAINER (panel->priv->a11y_button), image);
+        gtk_widget_show (image);
+        gtk_widget_show (panel->priv->a11y_button);
+        g_signal_connect (G_OBJECT (panel->priv->a11y_button),
+                          "clicked",
+                          G_CALLBACK (on_a11y_button_clicked), panel);
+
+        gtk_box_pack_start (GTK_BOX (panel->priv->hbox), panel->priv->a11y_button, FALSE, FALSE, 0);
+
+        panel->priv->language_option_widget = gdm_language_option_widget_new ();
+        g_signal_connect (G_OBJECT (panel->priv->language_option_widget),
+                          "language-activated",
+                          G_CALLBACK (on_language_activated), panel);
+        gtk_box_pack_start (GTK_BOX (panel->priv->hbox), panel->priv->language_option_widget, FALSE, FALSE, 6);
+
+        panel->priv->session_option_widget = gdm_session_option_widget_new ();
+        g_signal_connect (G_OBJECT (panel->priv->session_option_widget),
+                          "session-activated",
+                          G_CALLBACK (on_session_activated), panel);
+        gtk_box_pack_start (GTK_BOX (panel->priv->hbox), panel->priv->session_option_widget, FALSE, FALSE, 6);
 
         /* FIXME: we should only show hostname on panel when connected
            to a remote host */
@@ -412,11 +522,17 @@ gdm_greeter_panel_init (GdmGreeterPanel *panel)
                 gtk_widget_show (panel->priv->hostname_label);
         }
 
+        panel->priv->clock = gdm_clock_widget_new ();
+        gtk_box_pack_end (GTK_BOX (panel->priv->hbox),
+                          GTK_WIDGET (panel->priv->clock), FALSE, FALSE, 6);
+        gtk_widget_show (panel->priv->clock);
+
         tray = na_tray_new_for_screen (gtk_window_get_screen (GTK_WINDOW (panel)),
                                        GTK_ORIENTATION_HORIZONTAL);
         gtk_box_pack_end (GTK_BOX (panel->priv->hbox), GTK_WIDGET (tray), FALSE, FALSE, 6);
         gtk_widget_show (GTK_WIDGET (tray));
 
+        gdm_greeter_panel_hide_user_options (panel);
 }
 
 static void
@@ -444,3 +560,69 @@ gdm_greeter_panel_new (void)
 
         return GTK_WIDGET (object);
 }
+
+void
+gdm_greeter_panel_show_user_options (GdmGreeterPanel *panel)
+{
+        gtk_widget_show (panel->priv->session_option_widget);
+        gtk_widget_show (panel->priv->language_option_widget);
+}
+
+void
+gdm_greeter_panel_hide_user_options (GdmGreeterPanel *panel)
+{
+        gtk_widget_hide (panel->priv->session_option_widget);
+        gtk_widget_hide (panel->priv->language_option_widget);
+}
+
+void
+gdm_greeter_panel_reset (GdmGreeterPanel *panel)
+{
+        gdm_greeter_panel_set_default_language_name (panel, NULL);
+        gdm_greeter_panel_set_default_session_name (panel, NULL);
+        gdm_greeter_panel_hide_user_options (panel);
+}
+
+void
+gdm_greeter_panel_set_default_language_name (GdmGreeterPanel *panel,
+                                             const char      *language_name)
+{
+        char *normalized_language_name;
+
+        g_return_if_fail (GDM_IS_GREETER_PANEL (panel));
+
+        if (language_name != NULL) {
+                normalized_language_name = gdm_normalize_language_name (language_name);
+        } else {
+                normalized_language_name = NULL;
+        }
+
+        if (normalized_language_name != NULL &&
+            !gdm_option_widget_lookup_item (GDM_OPTION_WIDGET (panel->priv->language_option_widget),
+                                            normalized_language_name, NULL, NULL, NULL)) {
+                gdm_recent_option_widget_add_item (GDM_RECENT_OPTION_WIDGET (panel->priv->language_option_widget),
+                                                   normalized_language_name);
+        }
+
+        gdm_option_widget_set_default_item (GDM_OPTION_WIDGET (panel->priv->language_option_widget),
+                                            normalized_language_name);
+
+        g_free (normalized_language_name);
+}
+
+void
+gdm_greeter_panel_set_default_session_name (GdmGreeterPanel *panel,
+                                            const char      *session_name)
+{
+        g_return_if_fail (GDM_IS_GREETER_PANEL (panel));
+
+        if (session_name != NULL &&
+            !gdm_option_widget_lookup_item (GDM_OPTION_WIDGET (panel->priv->session_option_widget),
+                                            session_name, NULL, NULL, NULL)) {
+                g_warning ("Default session is not available");
+                return;
+        }
+
+        gdm_option_widget_set_default_item (GDM_OPTION_WIDGET (panel->priv->session_option_widget),
+                                            session_name);
+}
